%% simple region growing
%
%% input:
% depth - depth image
% mask - masks where the regions should be searched
% p - starting point(s), if several, of size Nx2

%% options:
% thrSpace - threashold in space (default 3)
% useAverage - use average distance (default false)
% useNormals - use normals (defaut false)
% normals - if use normals -> normals
% thrN - if use normals -> threashold on normals (default 1)
% idxImage - if use normals -> index

%% output:
% segmented region

function region = regionGrowing(depth,mask,p,options)

    region = zeros(size(depth,1),size(depth,2));
    region(sub2ind(size(region),p(:,1),p(:,2))) = 1;
    
    % process options
    if(isfield(options,'thrSpace'))
        thrSpace = options.thrSpace;
    else
        thrSpace = 3;
    end
    
    if(isfield(options,'maxSize'))
        con = options.maxSize;
    else
        con = -1;
    end
    
    if(isfield(options,'useAverage'))
        useAverage = options.useAverage;
    else
        useAverage = 0;
    end
    
    if (isfield(options,'useNormals'))
        useNormals = options.useNormals;
        if (useNormals == 1)
            if (isfield(options,'normals') && isfield(options,'idxImage'))
                normals = options.normals; %normals as an image with 3 components
                idxImg = options.idxImage;
                if (isfield(options,'thrN'))
                    thrN = options.thrN;
                else
                    thrN = 1;
                end
            else
                useNormals = 0;
                warning('Normals for pc are not specified; no normals will be used');
            end
        end
    else
        useNormals = 0;
    end
            
    fixedval = depth(p(1,1),p(1,2));
    
     while ~isempty(p)
        [nx,ny,nx0,ny0] = getNeighbours8(p,[size(depth,1) size(depth,2)]);
        
        
        if (0)
            %here is the problem - we dont necessarily choose the smallest distance
            
            [C,ia,~] = unique([nx ny],'rows','first');
        
            nx = C(:,1);
            ny = C(:,2);
        
            nx0 = nx0(ia);
            ny0 = ny0(ia);
            
        end
        
        linind = sub2ind(size(depth),nx,ny);
        
        nx(region(linind) == 1 | mask(linind) == 0) = []; 
        ny(region(linind) == 1 | mask(linind) == 0) = []; 
        nx0(region(linind) == 1 | mask(linind) == 0) = []; 
        ny0(region(linind) == 1 | mask(linind) == 0) = []; 
        
        linind = sub2ind(size(depth),nx,ny);
        
        if(con>0)
            nx(abs(depth(linind) - fixedval) > con) = [];
            ny(abs(depth(linind) - fixedval) > con) = [];
            nx0(abs(depth(linind) - fixedval) > con) = [];
            ny0(abs(depth(linind) - fixedval) > con) = [];
        end
        
        linind = sub2ind(size(depth),nx,ny);

        if(useAverage == 0)
            linind0 = sub2ind([size(depth,1) size(depth,2)],nx0,ny0);

            %compute space distance
            deltaSpace = abs(depth(linind) - depth(linind0));
            
            toRemove = (deltaSpace>thrSpace);
            
            %compute difference in normals
            if(useNormals == 1)
                %replace this distance with angle?
                deltaN = normals(idxImg(linind),:)-normals(idxImg(linind0),:);
                deltaN = sqrt(sum(deltaN.^2,2));
                
                toRemove = toRemove|(deltaN>thrN);
            end
            
            nx(toRemove) = []; 
            ny(toRemove) = [];     
        else
            avSpace = sum(sum(depth(region==1)))/sum(sum(region));
            
            %compute space distance
            deltaSpace = depth(linind) - avSpace;
            
            toRemove = deltaSpace>thrSpace;
            
            %compute difference in normals
            if(useNormals == 1)
                avN = mean(normals(idxImg(region==1),:),1);
                %replace this distance with angle?
                deltaN = normals(idxImg(linind),:) - repmat(avN,length(linind),1);
                deltaN = sqrt(sum(deltaN.^2,2));
                
                toRemove = toRemove|(deltaN>thrN);
            end
            
            nx(toRemove) = [];
            ny(toRemove) = [];
        end
        
        linind = sub2ind(size(depth),nx,ny);
        
        region(linind) = 1;
        
        if (~isempty(nx))
            C = unique([nx ny],'rows');
            nx = C(:,1); ny = C(:,2);
        end
        
        p = [nx ny];
    end

end

function [nx,ny,nx0,ny0] = getNeighbours8(p,dims)
    nx0 = repmat(p(:,1),8,1);
    ny0 = repmat(p(:,2),8,1);

    nx = [p(:,1)+1; p(:,1)-1; p(:,1); p(:,1); p(:,1)+1; p(:,1)+1; p(:,1)-1; p(:,1)-1];
    ny = [p(:,2); p(:,2); p(:,2)+1; p(:,2)-1; p(:,2)+1; p(:,2)-1; p(:,2)+1; p(:,2)-1];
    
    ny(nx<=0 | nx > dims(1)) = [];
    nx0(nx<=0 | nx > dims(1)) = [];
    ny0(nx<=0 | nx > dims(1)) = [];
    nx(nx<=0 | nx > dims(1)) = [];
    
    nx(ny<=0 | ny > dims(2)) = [];
    nx0(ny<=0 | ny > dims(2)) = [];
    ny0(ny<=0 | ny > dims(2)) = [];
    ny(ny<=0 | ny > dims(2)) = [];
end